(**************************************************************)
(*
 *  Ensemble, (Version 0.61)
 *  Copyright 1999 Cornell University
 *  All rights reserved.
 *
 *  See ensemble/doc/license.txt for further information.
 *)
(**************************************************************)
(**************************************************************)
(* UDP.ML *)
(* Author: Mark Hayden, 5/95 *)
(**************************************************************)
open Util
open Trans
(**************************************************************)
let name = Trace.file "UDP"
let failwith s = Trace.make_failwith name s
let log = Trace.log name
(**************************************************************)
(* NOTE: What about using an extra socket set to
 * no-local-receipt for sending multicasts on?  *)
(**************************************************************)
(* NOTE: This module uses separate sockets for sending and
 * receiving.  This is to avoid problems with Linux where
 * operations after sending to an unconnected port result in
 * ECONNREFUSED errors.  *)
(**************************************************************)

(* This buffer size was suggested by Werner Vogels.
 *)
let buffer_size = 64*1024*1024/(2*128+1024) (* = 52428 *)

(**************************************************************)

let init_xmit_sock () =
  let sock =
    try Hsys.socket_dgram () with e ->
      eprintf "UDP:error creating socket:%s, exiting\n" (Hsys.error e) ;
      exit 1
  in
  
  (* Set the socket up to be nonblocking.
   *)
  begin try
    Hsys.setsockopt sock (Hsys.Nonblock true) ;
  with e ->
    log (fun () -> sprintf "warning:setsockopt:Nonblock:%s" (Hsys.error e))
  end ;

  (* Try to set the size of the receive buffer.
   *)
  begin try
    Hsys.setsockopt sock (Hsys.Sendbuf buffer_size) ;
  with e ->
    log (fun () -> sprintf "warning:setsockopt:Sendbuf(%d):%s" buffer_size (Hsys.error e))
  end ;

  (* Bind to any port on all the interfaces.
   *)
  ignore (Hsys.bind_any "UDP:xmit" sock (Hsys.inet_any ())) ;

  sock

(**************************************************************)

let init_recv_sock host =
  let sock =
    try Hsys.socket_dgram () with e ->
      eprintf "UDP:error creating socket:%s, exiting\n" (Hsys.error e) ;
      exit 1
  in
    
  begin 
    begin try
      (* Set the socket to be nonblocking.
       *)
      Hsys.setsockopt sock (Hsys.Nonblock true) ;
    with e ->
      log (fun () -> sprintf "warning:setsockopt:Nonblock:%s" (Hsys.error e))
    end ;

    (* Try to set the size of the receive buffer.
     *)
    begin try
      Hsys.setsockopt sock (Hsys.Recvbuf buffer_size) ;
    with e ->
      log (fun () -> sprintf "warning:setsockopt:Recvbuf(%d):%s" buffer_size (Hsys.error e))
    end ;
  end ;

  (* Set the socket to close on exec.
   *)
  (*set_close_on_exec sock ;*)

  (* Bind to any port and the specified address.
   *)
  let port = Hsys.bind_any "UDP:recv" sock host in
  log (fun () -> sprintf "kernel bound port %d (receiving)" port) ;

  (sock,port)

(**************************************************************)

let domain alarm =
  let handlers = Alarm.handlers alarm in
  (* Note that deering port is always decided by the
   * configuration.  We should be able to bind to this port
   * without problems because all Ensemble processes use
   * reusable IP multicast sockets.  
   *)
  let deering_port () = Arge.check name Arge.deering_port in
  let gossip_port = do_once (fun () -> Arge.check name Arge.gossip_port) in
  let gossip_hosts = do_once (fun () ->
      let hosts = Arge.check name Arge.gossip_hosts in
      let hosts = Arge.inet_of_string_list Arge.gossip_hosts hosts in
      Array.of_list hosts
    )
  in

  (* This allows the user to override the host name to use to
   * send messages to me.
   *)
  let host =
    match Arge.get Arge.udp_host with
    | None -> Hsys.gethost ()
    | Some host -> Arge.inet_of_string Arge.udp_host host
  in
(*
  let map_localhost = ident in
*)

  let localhost = Hsys.getlocalhost() in
  let map_localhost inet =
    if inet = host then localhost else inet
  in

  let sp2_info = do_once (fun () ->
    let host = Hsys.gethostname () in
    let try_suffix suffix =
      try
	match string_split "." host with
	| [p1;p2;p3;p4] ->
	    let p1 = p1 ^ suffix in
	    let host = [p1;p2;p3;p4] in
	    let host = String.concat "." host in
	    eprintf "UDP:trying sp2 udp hostname:'%s'\n" host ;
	    let host = Hsys.inet_of_string host in
	    Some(host)
	| _ -> failwith "bad inet list"
      with _ ->
	None
    in
    let suffixes = Arge.get Arge.sp2_suffixes in
    let rec loop = function
      | [] -> failwith "no suffixes worked for sp2 fast interconnect hostname"
      | hd :: tl ->
	  match try_suffix hd with
	  | None -> loop tl
	  | Some inet ->
	      let (sock,port) = init_recv_sock inet in
	      let addr = Addr.Sp2A(inet,port) in
	      (sock,addr)
    in loop suffixes
  ) in

  (* Initialize send and receive sockets.  Both are bound
   * to INADDR_ANY so that messages will be accepted on
   * all network interfaces.
   *)
  let (recv_sock, port) = init_recv_sock (Hsys.inet_any ()) in
  let xmit_sock = init_xmit_sock () in

  let xmit_sock_mc = do_once (fun () ->
    if not (Hsys.has_ip_multicast()) then
      failwith "multicast not supported" ;
    let sock = init_xmit_sock () in
    Hsys.setsockopt sock (Hsys.Multicast true) ;
    sock
  )
  in

  (* Tell the Appl module our unique port number.
   *)
  Appl.install_port port ;

  (* TODO: I'm having problems with the loopback stuff.
   *)
  let noloop = xmit_sock in
(*
  let noloop = socket_dgram () in
  setsock_multicast noloop false ;
*)

  let udp_addr = Addr.UdpA(host,port) in

  (* Note: we give out the same port as for UDP.  This
   * is for sending messages pt2pt to me.
   *)
  let deering_addr () = Addr.DeeringA(host,port) in

  let addr mode = match mode with
  | Addr.Udp     -> udp_addr
  | Addr.Deering -> deering_addr ()
  | Addr.Sp2     -> snd (sp2_info ())
  | _ -> failwith "addr:bad mode"
  in

  let xmit mode dest =
    (* Do some preprocessing.
     *)
    let dests =
      match dest with
      | Domain.Pt2pt dests ->
	  if Arrayf.length dests = 0 then 
	    None
	  else (
	    (* Depending on the type of the address, use
	     * either the standard Udp socket or the
	     * special Sp2 socket.  
	     *)
	    let sock = match mode with
	    | Addr.Udp -> xmit_sock
	    | Addr.Deering -> xmit_sock_mc()
	    | Addr.Sp2 -> fst (sp2_info ())
	    | _ -> failwith "xmit:sanity"
	    in

	    let dests =
	      Arrayf.map (fun dest ->
		match Addr.project dest mode with 
		| Addr.UdpA    (inet,port) -> (map_localhost inet,port)
		| Addr.DeeringA(inet,port) -> (map_localhost inet,port)
		| Addr.Sp2A    (inet,port) -> (map_localhost inet,port)
		| _ -> failwith "xmit:sanity"
	      ) dests
	    in
	    let dests = Arrayf.to_array dests in
	    Some (Hsys.sendto_info sock dests)
	  )
      | Domain.Mcast(dest,loopback) ->
	  (* Multicast communication mean that this must
	   * be a Deering address.
           *)
	  (*
	    let sock = if loopback then xmit_sock else noloop in
	   *)
	  let sock = xmit_sock_mc () in
	  let hash = Group.hash_of_id dest in
	  let inet = Hsys.deering_addr hash in 
	  Some (Hsys.sendto_info sock [|inet,deering_port()|])
      | Domain.Gossip(dest) ->
      	  match mode with
	    (* For gossip messages with Deering, just broadcast
             * it as above.
	     *)
	  | Addr.Deering ->
	      let hash = Group.hash_of_id dest in
	      let inet = Hsys.deering_addr hash in 
	      Some (Hsys.sendto_info (xmit_sock_mc()) [|inet,deering_port()|])

	    (* For Udp, we send the message to the various 
	     * gossip hosts.
	     *)
	  | Addr.Udp ->
	      (* For this, we need to send using the recv_sock because
	       * the gossip server will use recvfrom to learn the
	       * address to reply to.
	       *)
	      let gossip_port = gossip_port () in
	      let dests = Array.map (fun host -> (host,gossip_port)) (gossip_hosts()) in

	      (* HACK for Linux. I also put my address in
	       * the list of destinations.  On Linux, this
	       * clears the error in the socket, if there is
	       * one.  
	       *)
	      let dests = Array.append dests [|host,gossip_port|] in

	      if dests = [||] then None else
	        Some (Hsys.sendto_info recv_sock dests)
	  | _ -> failwith "bad mode"
    in

    option_map (fun dests ->
      let x buf ofs len = Buf.sendto dests buf ofs len in
      let xv iov = Iovecl.sendtov dests iov ; Iovecl.free name iov in
      let xvs iov s = Iovecl.sendtovs dests iov s ; Iovecl.free name iov in
      let x,xv,xvs =
	if Arge.timestamp_check "send" then (
	  let ts_xmit = Elink.get name Elink.timestamp_add "UDP:xmit" in
	  let x buf ofs len = ts_xmit () ; x buf ofs len
	  and xv iov = ts_xmit () ; xv iov
	  and xvs iov s = ts_xmit () ; xvs iov s
	  in (x,xv,xvs)
	) else (x,xv,xvs)
      in (x,xv,xvs)
    ) dests
  in

  let enable mode group endpt view =
    match mode with
    | Addr.Udp ->
    	Alarm.add_sock_recv alarm name recv_sock Hsys.Handler1
    | Addr.Deering ->
      	let hash = Group.hash_of_id group in
      	let deering_sock = Ipmc.join (Hsys.deering_addr hash) (deering_port ()) in
    	Alarm.add_sock_recv alarm name recv_sock Hsys.Handler1 ;
      	Alarm.add_sock_recv alarm "DEERING" deering_sock Hsys.Handler1
    | Addr.Sp2 ->
	let sp2_sock = fst (sp2_info ()) in
    	Alarm.add_sock_recv alarm "SP2" sp2_sock Hsys.Handler1
    | _ -> failwith "enable:bad mode"
  in

  let disable mode group endpt view =
    match mode with
    | Addr.Udp ->
    	Alarm.rmv_sock_recv alarm recv_sock
    | Addr.Deering ->
      	let hash = Group.hash_of_id group in
      	let deering_sock = Ipmc.leave (Hsys.deering_addr hash) (deering_port()) in
    	Alarm.rmv_sock_recv alarm recv_sock ;
      	Alarm.rmv_sock_recv alarm deering_sock
    | Addr.Sp2 ->
	let sp2_sock = fst (sp2_info ()) in
    	Alarm.rmv_sock_recv alarm sp2_sock
    | _ -> failwith "disable:bad mode"
  in 

  Domain.create name addr enable disable xmit

(**************************************************************)

let _ = Elink.put Elink.udp_domain domain

(**************************************************************)
