(**************************************************************)
(*
 *  Ensemble, (Version 0.61)
 *  Copyright 1999 Cornell University
 *  All rights reserved.
 *
 *  See ensemble/doc/license.txt for further information.
 *)
(**************************************************************)
(**************************************************************)
(* EVENT.MLI *)
(* Author: Mark Hayden, 4/95 *)
(* Based on Horus events by Robbert vanRenesse *)
(**************************************************************)
open Trans
(**************************************************************)

  (* Up and down events *)
type t
type up = t
type dn = t

(**************************************************************)
(* 
 * ACKNOWLEDGEMENT: fields of this type appear in both up
 * and down events.  The default value for an acknowledgement
 * is NoAck.  The usual way acknowledgements work is for
 * some layer to set the ack field in an up event with an
 * acknowledgement that it wants the event to be
 * acknowledged with.  The layer which initiate the message
 * then is responsible for later delivering down a EAck with
 * the ack field set to be the same acknowledgment.  
 *)
(*
type acknowledgement =
  | NoAck
  | RankSeqno of origin * seqno

val string_of_ack : acknowledgement -> string
*)
(**************************************************************)

  (* Event types *)
type typ =
    (* These events should have messages associated with them. *)
  | ECast				(* Multicast message *)
  | ESend				(* Pt2pt message *)
  | ESubCast				(* Multi-destination message *)
  | ECastUnrel				(* Unreliable multicast message *)
  | ESendUnrel				(* Unreliable pt2pt message *)
  | EMergeRequest			(* Request a merge *)
  | EMergeGranted			(* Grant a merge request *)
  | EOrphan				(* Message was orphaned *)

    (* These types do not have messages. *)
  | EAccount				(* Output accounting information *)
(*| EAck			      *)(* Acknowledge message *)
  | EAsync				(* Asynchronous application event *)
  | EBlock				(* Block the group *)
  | EBlockOk				(* Acknowledge blocking of group *)
  | EDump				(* Dump your state (debugging) *)
  | EElect				(* I am now the coordinator *)
  | EExit				(* Disable this stack *)
  | EFail				(* Fail some members *)
  | EGossipExt				(* Gossip message *)
  | EInit				(* First event delivered *)
  | ELeave				(* A member wants to leave *)
  | ELostMessage			(* Member doesn't have a message *)
  | EMergeDenied			(* Deny a merge request *)
  | EMergeFailed			(* Merge request failed *)
  | EMigrate				(* Change my location *)
  | EPresent      (* Members present in this view *)
  | EPrompt				(* Prompt a new view *)
  | EProtocol				(* Request a protocol switch *)
  | ERekey				(* Request a rekeying of the group *)
  | EStable				(* Deliver stability down *)
  | EStableReq				(* Request for stability information *)
  | ESuspect				(* Member is suspected to be faulty *)
  | ESystemError			(* Something serious has happened *)
  | ETimer				(* Request a timer *)
  | EView				(* Notify that a new view is ready *)
  | EXferDone				(* Notify that a state transfer is complete *)
  | ESyncInfo
  | ESecureMsg				(* Ohad -- Private Secure messaging *)
  | EChannelList			(* Ohad -- passing a list of secure-channels *)
  | EMflowBlock				(* Ohad -- Blocking/unblocking the application for mflow control*)

(**************************************************************)

type field =
      (* Common fields *)
  | Type	of typ			(* type of the message*)
  | Peer	of rank			(* rank of sender/destination *)
(*| Ack	        of acknowledgement    *)(* acknowledgement information *)
  | Iov	        of Iovecl.t		(* payload of message *)
  | ApplMsg				(* was this message generated by an appl? *)

      (* Uncommon fields *)
  | Address     of Addr.set		(* new address for a member *)
  | Failures	of bool Arrayf.t	(* failed members *)
  | Presence    of bool Arrayf.t        (* members present in the current view *)
  | Suspects	of bool Arrayf.t        (* suspected members *)
  | SuspectReason of string		(* reasons for suspicion *)
  | Stability	of seqno Arrayf.t	(* stability vector *)
  | NumCasts	of seqno Arrayf.t	(* number of casts seen *)
  | Mergers	of View.state		(* list of merging members *)
  | Contact	of Endpt.full * View.id option (* contact for a merge *)
  | HealGos	of Proto.id * View.id * Endpt.full * View.t (* HEAL gossip *)
  | SwitchGos	of Proto.id * View.id * Time.t  (* SWITCH gossip *)
  | ExchangeGos	of string		(* EXCHANGE gossip *)
  | ViewState	of View.state		(* state of next view *)
  | ProtoId	of Proto.id		(* protocol id (only for down events) *)
  | Time        of Time.t		(* current time *)
  | Alarm       of Time.t		(* for alarm requests *)
  | ApplCasts   of seqno Arrayf.t
  | ApplSends   of seqno Arrayf.t
  | DbgName     of string

      (* Flags *)
  | NoTotal				(* message is not totally ordered*)
  | ServerOnly				(* deliver only at servers *)
  | ClientOnly				(* deliver only at clients *)
  | NoVsync
  | Fragment				(* Iovec has been fragmented *)

      (* Debugging *)
  | History	of string		(* debugging history *)

      (* Ohad -- Private Secure Messaging *)
  | SecureMsg of (rank * Buf.t)
  | ChannelList of (rank * Security.key) list
	
      (* Ohad -- interaction between Mflow and the application*)
  | MflowBlock of bool

(**************************************************************)

  (* Directed events. *)
type dir = Up of t | Dn of t

type ('a,'b) dirm = UpM of up * 'a | DnM of dn * 'b

(**************************************************************)

  (* Constructor *)
val create	: debug -> typ -> field list -> t

(**************************************************************)

  (* Modifier *)
val set		: debug -> t -> field list -> t

(**************************************************************)

  (* Copier *) 
val copy	: debug -> t -> t

(**************************************************************)

  (* Destructor *)
val free	: debug -> t -> unit

(* Same as above, but don't modify iovec refcounts.
 *)
val free_noIov	: debug -> t -> unit

(**************************************************************)

  (* Sanity checkers *)
val upCheck	: debug -> t -> unit
val dnCheck	: debug -> t -> unit

(**************************************************************)

  (* Pretty printer *)
val to_string	: t -> string
val string_of_type : typ -> string

(**************************************************************)

  (* Special constructors *)

val bodyCore		: debug -> typ -> origin -> Iovecl.t -> t
(*val bypassCast	: debug -> origin -> Iovecl.t -> t*)
(*val bypassSend	: debug -> origin -> Iovecl.t -> t*)
val castEv		: debug	-> t
val castIov		: debug -> Iovecl.t -> t
val castIovAppl		: debug -> Iovecl.t -> t
val castPeerIov 	: debug -> rank -> Iovecl.t -> t
val sendPeer		: debug -> rank	-> t
val sendPeerIov         : debug -> rank -> Iovecl.t -> t
val sendPeerIovAppl	: debug -> rank -> Iovecl.t -> t
val sendUnrelPeer       : debug -> rank -> t
val suspectReason       : debug -> bool Arrayf.t -> debug -> t
val timerAlarm		: debug -> Time.t -> t
val timerTime		: debug -> Time.t -> t
val dumpEv              : debug -> t

(**************************************************************)

  (* Accessors *)

(*val getAck		: t -> acknowledgement*)
val getAddress     	: t -> Addr.set
val getAlarm		: t -> Time.t
val getApplMsg	        : t -> bool
val getApplCasts	: t -> seqno Arrayf.t
val getApplSends  	: t -> seqno Arrayf.t
val getClientOnly	: t -> bool
val getContact		: t -> Endpt.full * View.id option
val getExtend     	: t -> field list
val getExtender     	: (field -> 'a option) -> 'a -> t -> 'a
val getExtendOpt	: t -> (field -> bool) -> unit
val getFailures		: t -> bool Arrayf.t
val getFragment		: t -> bool
val getIov		: t -> Iovecl.t
val getMergers		: t -> View.state
val getNoTotal		: t -> bool
val getNoVsync          : t -> bool
val getNumCasts		: t -> seqno Arrayf.t
val getPeer		: t -> origin
val getDbgName          : t -> string
val getPresence         : t -> bool Arrayf.t
val getProtoId		: t -> Proto.id
val getServerOnly	: t -> bool
val getStability	: t -> seqno Arrayf.t
val getSuspectReason	: t -> string
val getSuspects		: t -> bool Arrayf.t
val getTime		: t -> Time.t
val getType		: t -> typ
val getViewState	: t -> View.state
val getSecureMsg        : t -> rank * Buf.t
val getChannelList      : t -> (Trans.rank * Security.key) list
val getMflowBlock       : t -> bool	

val setIovFragment	: debug -> t -> Iovecl.t -> t
val setNoTotal  	: debug -> t -> t

(**************************************************************)

val getIovLen           : t -> Buf.len

(**************************************************************)

  (* For automating acknowledgements *)
(*val make_acker	: string -> (t -> unit) -> t -> unit*)

(**************************************************************)
