(**************************************************************)
(*
 *  Ensemble, (Version 0.61)
 *  Copyright 1999 Cornell University
 *  All rights reserved.
 *
 *  See ensemble/doc/license.txt for further information.
 *)
(**************************************************************)
(**************************************************************)
(* TIME.ML *)
(* Author: Mark Hayden, 8/96 *)
(**************************************************************)
open Util
open Hsys
(**************************************************************)
let name = Trace.file "TIME"
let failwith = Trace.make_failwith name
(**************************************************************)

type t = Hsys.timeval

let million = 1000000

let rec of_ints sec usec = 
  if usec < 0 then (
    of_ints (pred sec) (usec + million)
  ) else if usec >= million then (
    let sec = sec + usec / million in
    let usec = usec mod million in
    assert (usec >= 0 && usec < million) ;
    {sec=sec;usec=usec}
  ) else {sec=sec;usec=usec}

let of_ints sec usec = 
  if usec < 0 || usec >= million then
    of_ints sec usec
  else {sec=sec;usec=usec}
  
let of_int sec = of_ints sec 0

let zero = {sec=0;usec=0}
let neg_one = {sec=(-1);usec=0}
let invalid = neg_one

let to_timeval {sec=sec;usec=usec} = {sec=sec;usec=usec}

(* Print out time with millisecond accuracy
 *)
let to_string t = 
  sprintf "%d.%03d" t.sec (t.usec/1000)

(* Convert with full accuracy.
 *)
let to_string_full t = 
  sprintf "%d.%06d" t.sec t.usec

let add a b =
  of_ints (a.sec + b.sec) (a.usec + b.usec)

let sub a b = 
  of_ints (a.sec - b.sec) (a.usec - b.usec)

let of_float f =
  let sec = floor f in
  let usec = (f -. sec) *. 1.0E6 in
  let sec = truncate sec in
  let usec = truncate usec in
  of_ints sec usec

let of_string s =
(*
  eprintf "TIME:of_string:%s\n" s ;
*)
  try
    let point = strchr '.' s in
    let sec = 
      if point = 0 then 
	0 
      else
      	int_of_string (String.sub s 0 point) 
    in

    let usec = 
      let usec = String.sub s (succ point) (String.length s - point - 1) in
      let useclen = String.length usec in
      if useclen = 0 then 0 else (
	let usec = int_of_string usec in
	let rec loop usec len =
	  if len < 6 then
	    loop (usec * 10) (succ len)
	  else if len > 6 then
	    loop (usec / 10) (pred len)
	  else usec
	in
	let usec = loop usec useclen in
	usec
      )
    in
    of_ints sec usec
  with Not_found ->
    of_int (int_of_string s)
(*
let of_string s =
  let time = float_of_string s in
  let time = of_float time in
  let new_time = of_string s in
  if new_time.sec <> time.sec 
  || abs (new_time.usec - time.usec) > 10 
  then
    failwith "new time is bad" ;
  new_time
*)

let to_float tv =
  (float tv.sec) +. (float tv.usec /. 1.0E6)

let is_zero t = t.usec = 0 && t.sec = 0

let ge a b =
  if a.sec = b.sec then
    a.usec >= b.usec
  else 
    a.sec > b.sec

let gt a b =
  if a.sec = b.sec then
    a.usec > b.usec
  else 
    a.sec > b.sec

module Ord =
  struct
    type time = t
    type t = time
    let zero = zero
    let ge = ge
  end

let gettimeofday = Hsys.gettimeofday
let select = Hsys.select

(**************************************************************)
(* Another version of the above stuff with in-place 
 * modification.  Use at your own risk.
 *)
let rec set t sec usec = 
  if usec >= 0 && usec < million then (
    t.sec <- sec ; 
    t.usec <- usec
  ) else if usec < 0 then (
    set t (pred sec) (usec + million)
  ) else (
    set t (succ sec) (usec - million)
  )

type m = t
let mut_copy {sec=sec;usec=usec} = {sec=sec;usec=usec}
let mut () = mut_copy zero
let mut_set m t = m.sec <- t.sec ; m.usec <- t.usec
let mut_gettimeofday = Hsys.gettimeofdaya
let mut_ge = ge
let mut_add m t = set m (m.sec + t.sec) (m.usec + t.usec)
let mut_sub m t = set m (m.sec - t.sec) (m.usec - t.usec)
let mut_sub_rev t m = set m (t.sec - m.sec) (t.usec - m.usec)
let mut_select = Hsys.select

(*
type m' = m
module Mut_hide = struct
  type m = m'
  let copy = mut_copy
  let zero = mut_zero
  let gettimeofday = mut_gettimeofday
  let ge = mut_ge
  let of_int = mut_of_int
  let add = mut_add
  let sub = mut_sub
  let sub_rev = mut_sub_rev
  let select = mut_select
end
*)
(**************************************************************)
